@@ -11,6 +11,13 @@ postgresql (usually, the system wide python installation).
 """
 
 import sys
+from collections import namedtuple
+try:
+    from collections import OrderedDict
+except ImportError:
+    import collections
+    from ordereddict import OrderedDict
+    collections.OrderedDict = OrderedDict
 
 __version__ = '__VERSION__'
 
@@ -19,6 +26,29 @@ ALL = object()
 UNBOUND = object()
 
 
+SortKey = namedtuple("SortKey",
+                              ["attname", "attnum", "is_reversed",
+                               "nulls_first", "collate"])
+
+"""
+A SortKey describes the sort of one column an SQL query requested.
+
+A query can request the sort of zero, one or multiple columns. Therefore, a list
+of SortKey is provided to the ForeignDataWrapper, containing zero, one or more
+SortKey.
+
+Attributes:
+    attname(str):       The name of the column to sort as defined in the postgresql
+        table.
+    attnum(int):        The position of the column to sort as defined in the
+        postgresql table.
+    is_reversed(bool):  True is the query requested a DESC order.
+    nulls_first(bool):  If True, NULL values must appears at the beginning.
+        Otherwise, they must appear at the end.
+    collate(str):       The collation name to use to sort the data, as appearing
+        in the postgresql cluster.
+"""
+
 class Qual(object):
     """A Qual describes a postgresql qualifier.
 
@@ -152,6 +182,32 @@ class ForeignDataWrapper(object):
         """
         return (100000000, len(columns) * 100)
 
+    def can_sort(self, sortkeys):
+        """
+        Method called from the planner to ask the FDW what are the sorts it can
+        enforced, to avoid PostgreSQL to sort the data after retreiving all the
+        rows. These sorts can come from explicit ORDER BY clauses, but also GROUP
+        BY and DISTINCT clauses.
+
+        The FDW has to inspect every sort, and respond which one are handled.
+        The sorts are cumulatives. For example::
+
+            col1 ASC
+            col2 DESC
+
+        means that the FDW must render the tuples sorted by col1 ascending and
+        col2 descending.
+
+        Args:
+            sortkeys (list): A list of :class:`SortKey`
+                representing all the sorts the query must enforce.
+
+        Return:
+            The list of cumulative SortKey, for which the FDW can
+            enforce the sort.
+        """
+        return []
+
     def get_path_keys(self):
         u"""
         Method called from the planner to add additional Path to the planner.
@@ -209,7 +265,18 @@ class ForeignDataWrapper(object):
         """
         return []
 
-    def execute(self, quals, columns):
+    def explain(self, quals, columns, sortkeys=None, verbose=False):
+        """Hook called on explain.
+
+        The arguments are the same as the :meth:`execute`, with the addition of
+        a "verbose" keyword arg for when the EXPLAIN is called with the VERBOSE
+        option.
+        Returns:
+            An iterable of strings to display in the EXPLAIN output.
+        """
+        return []
+
+    def execute(self, quals, columns, sortkeys=None):
         """Execute a query in the foreign data wrapper.
 
         This method is called at the first iteration.
@@ -240,6 +307,8 @@ class ForeignDataWrapper(object):
                 You should return AT LEAST those columns when returning a
                 dict. If returning a sequence, every column from the table
                 should be in the sequence.
+            sortkeys (list): A list of :class:`SortKey`
+                that the FDW said it can enforce.
 
         Returns:
             An iterable of python objects which can be converted back to PostgreSQL.
@@ -247,6 +316,8 @@ class ForeignDataWrapper(object):
             - sequences containing exactly as much columns as the
             underlying tables
             - dictionaries mapping column names to their values.
+            If the sortkeys wasn't empty, the FDW has to return the data in the
+            expected order.
 
         """
         pass
