@@ -4,6 +4,7 @@ from multicorn.compat import unicode_
 from .utils import log_to_postgres, WARNING, ERROR
 from itertools import cycle
 from datetime import datetime
+from operator import itemgetter
 
 
 class TestForeignDataWrapper(ForeignDataWrapper):
@@ -70,14 +71,31 @@ class TestForeignDataWrapper(ForeignDataWrapper):
                                                           index)
             yield line
 
-    def execute(self, quals, columns):
+    def execute(self, quals, columns, sortkeys=None):
+        sortkeys = sortkeys or []
         log_to_postgres(str(sorted(quals)))
         log_to_postgres(str(sorted(columns)))
+        if (len(sortkeys)) > 0:
+            log_to_postgres("requested sort(s): ")
+            for k in sortkeys:
+                log_to_postgres(k)
         if self.test_type == 'None':
             return None
         elif self.test_type == 'iter_none':
             return [None, None]
         else:
+            if (len(sortkeys) > 0):
+                # testfdw don't have tables with more than 2 fields, without
+                # duplicates, so we only need to worry about sorting on 1st
+                # asked column
+                k = sortkeys[0];
+                res = self._as_generator(quals, columns)
+                if (self.test_type == 'sequence'):
+                    return sorted(res, key=itemgetter(k.attnum - 1),
+                                  reverse=k.is_reversed)
+                else:
+                    return sorted(res, key=itemgetter(k.attname),
+                                  reverse=k.is_reversed)
             return self._as_generator(quals, columns)
 
     def get_rel_size(self, quals, columns):
@@ -90,6 +108,10 @@ class TestForeignDataWrapper(ForeignDataWrapper):
             return [(('test1',), 1)]
         return []
 
+    def can_sort(self, sortkeys):
+        # assume sort pushdown ok for all cols, in any order, any collation
+        return sortkeys
+
     def update(self, rowid, newvalues):
         if self.test_type == 'nowrite':
             super(TestForeignDataWrapper, self).update(rowid, newvalues)
