@@ -7,7 +7,6 @@
  * author: Kozea
  */
 #include "multicorn.h"
-#include "commands/explain.h"
 #include "optimizer/paths.h"
 #include "optimizer/pathnode.h"
 #include "optimizer/planmain.h"
@@ -22,6 +21,7 @@
 #include "catalog/pg_type.h"
 #include "utils/memutils.h"
 #include "miscadmin.h"
+#include "utils/lsyscache.h"
 #include "utils/rel.h"
 #include "parser/parsetree.h"
 
@@ -55,9 +55,12 @@ static ForeignScan *multicornGetForeignPlan(PlannerInfo *root,
 						Oid foreigntableid,
 						ForeignPath *best_path,
 						List *tlist,
-						List *scan_clauses);
-static void multicornExplainForeignScan(ForeignScanState *node,
-							ExplainState *es);
+						List *scan_clauses
+#if PG_VERSION_NUM >= 90500
+						, Plan *outer_plan
+#endif
+		);
+static void multicornExplainForeignScan(ForeignScanState *node, ExplainState *es);
 static void multicornBeginForeignScan(ForeignScanState *node, int eflags);
 static TupleTableSlot *multicornIterateForeignScan(ForeignScanState *node);
 static void multicornReScanForeignScan(ForeignScanState *node);
@@ -304,23 +307,71 @@ multicornGetForeignPaths(PlannerInfo *root,
 						 RelOptInfo *baserel,
 						 Oid foreigntableid)
 {
-	Path	   *path;
-	MulticornPlanState *planstate = baserel->fdw_private;
+	List				*pathes; /* List of ForeignPath */
+	MulticornPlanState	*planstate = baserel->fdw_private;
+	ListCell		    *lc;
+
+	/* These lists are used to handle sort pushdown */
+	List				*apply_pathkeys = NULL;
+	List				*deparsed_pathkeys = NULL;
 
 	/* Extract a friendly version of the pathkeys. */
 	List	   *possiblePaths = pathKeys(planstate);
 
-	findPaths(root, baserel, possiblePaths, planstate->startupCost);
-	/* Add a default path */
-	path = (Path *) create_foreignscan_path(root, baserel,
-											baserel->rows,
-											planstate->startupCost,
-											baserel->rows * baserel->width,
-											NIL,		/* no pathkeys */
-											NULL,		/* no outer rel either */
-											(void *) baserel->fdw_private);
+	/* Try to find parameterized paths */
+	pathes = findPaths(root, baserel, possiblePaths, planstate->startupCost,
+			planstate, apply_pathkeys, deparsed_pathkeys);
 
-	add_path(baserel, path);
+	/* Add a simple default path */
+	pathes = lappend(pathes, create_foreignscan_path(root, baserel,
+			baserel->rows,
+			planstate->startupCost,
+			baserel->rows * baserel->width,
+			NIL,		/* no pathkeys */
+		    NULL,
+#if PG_VERSION_NUM >= 90500
+			NULL,
+#endif
+			NULL));
+
+	/* Handle sort pushdown */
+	if (root->query_pathkeys)
+	{
+		List		*deparsed = deparse_sortgroup(root, foreigntableid, baserel);
+
+		if (deparsed)
+		{
+			/* Update the sort_*_pathkeys lists if needed */
+			computeDeparsedSortGroup(deparsed, planstate, &apply_pathkeys,
+					&deparsed_pathkeys);
+		}
+	}
+
+	/* Add each ForeignPath previously found */
+	foreach(lc, pathes)
+	{
+		ForeignPath *path = (ForeignPath *) lfirst(lc);
+
+		/* Add the path without modification */
+		add_path(baserel, (Path *) path);
+
+		/* Add the path with sort pusdown if possible */
+		if (apply_pathkeys && deparsed_pathkeys)
+		{
+			ForeignPath *newpath;
+
+			newpath = create_foreignscan_path(root, baserel, path->path.rows,
+					path->path.startup_cost, path->path.total_cost,
+					apply_pathkeys, NULL,
+#if PG_VERSION_NUM >= 90500
+					NULL,
+#endif
+					(void *) deparsed_pathkeys);
+
+			newpath->path.param_info = path->path.param_info;
+			add_path(baserel, (Path *) newpath);
+		}
+	}
 	errorCheck();
 }
 
@@ -334,7 +385,11 @@ multicornGetForeignPlan(PlannerInfo *root,
 						Oid foreigntableid,
 						ForeignPath *best_path,
 						List *tlist,
-						List *scan_clauses)
+						List *scan_clauses
+#if PG_VERSION_NUM >= 90500
+						, Plan *outer_plan
+#endif
+		)
 {
 	Index		scan_relid = baserel->relid;
 	MulticornPlanState *planstate = (MulticornPlanState *) baserel->fdw_private;
@@ -351,11 +406,18 @@ multicornGetForeignPlan(PlannerInfo *root,
 								&planstate->qual_list);
 		}
 	}
+	planstate->pathkeys = (List *) best_path->fdw_private;
 	return make_foreignscan(tlist,
 							scan_clauses,
 							scan_relid,
 							scan_clauses,		/* no expressions to evaluate */
-							serializePlanState(planstate));
+							serializePlanState(planstate)
+#if PG_VERSION_NUM >= 90500
+							, NULL
+							, NULL /* All quals are meant to be rechecked */
+							, NULL
+#endif
+							);
 }
 
 /*
@@ -367,6 +429,17 @@ multicornGetForeignPlan(PlannerInfo *root,
 static void
 multicornExplainForeignScan(ForeignScanState *node, ExplainState *es)
 {
+	PyObject *p_iterable = execute(node, es),
+			 *p_item,
+			 *p_str;
+	Py_INCREF(p_iterable);
+	while((p_item = PyIter_Next(p_iterable))){
+		p_str = PyObject_Str(p_item);
+		ExplainPropertyText("Multicorn", PyString_AsString(p_str), es);
+		Py_DECREF(p_str);
+	}
+	Py_DECREF(p_iterable);
+	errorCheck();
 }
 
 /*
@@ -416,7 +489,7 @@ multicornIterateForeignScan(ForeignScanState *node)
 
 	if (execstate->p_iterator == NULL)
 	{
-		execute(node);
+		execute(node, NULL);
 	}
 	ExecClearTuple(slot);
 	if (execstate->p_iterator == Py_None)
@@ -912,10 +985,13 @@ serializePlanState(MulticornPlanState * state)
 	List	   *result = NULL;
 
 	result = lappend(result, makeConst(INT4OID,
-						  -1, InvalidOid, -1, state->numattrs, false, true));
+						  -1, InvalidOid, 4, Int32GetDatum(state->numattrs), false, true));
 	result = lappend(result, makeConst(INT4OID,
-					-1, InvalidOid, -1, state->foreigntableid, false, true));
+					-1, InvalidOid, 4, Int32GetDatum(state->foreigntableid), false, true));
 	result = lappend(result, state->target_list);
+
+	result = lappend(result, serializeDeparsedSortGroup(state->pathkeys));
+
 	return result;
 }
 
@@ -930,10 +1006,13 @@ initializeExecState(void *internalstate)
 	List	   *values = (List *) internalstate;
 	AttrNumber	attnum = ((Const *) linitial(values))->constvalue;
 	Oid			foreigntableid = ((Const *) lsecond(values))->constvalue;
+	List		*pathkeys;
 
 	/* Those list must be copied, because their memory context can become */
 	/* invalid during the execution (in particular with the cursor interface) */
 	execstate->target_list = copyObject(lthird(values));
+	pathkeys = lfourth(values);
+	execstate->pathkeys = deserializeDeparsedSortGroup(pathkeys);
 	execstate->fdw_instance = getInstance(foreigntableid);
 	execstate->buffer = makeStringInfo();
 	execstate->cinfos = palloc0(sizeof(ConversionInfo *) * attnum);
