@@ -36,7 +36,8 @@ PyObject *pythonQual(char *operatorname, PyObject *value,
 		   bool use_or,
 		   Oid typeoid);
 
-
+PyObject  *getSortKey(MulticornDeparsedSortGroup *key);
+MulticornDeparsedSortGroup *getDeparsedSortGroup(PyObject *key);
 
 
 Datum pyobjectToDatum(PyObject *object, StringInfo buffer,
@@ -483,9 +484,9 @@ getColumnsFromTable(TupleDesc desc, PyObject **p_columns, List **columns)
 				errorCheck();
 				columnDef = lappend(columnDef, makeString(key));
 				columnDef = lappend(columnDef, makeConst(TYPEOID,
-								   -1, InvalidOid, -1, typOid, false, true));
-				columnDef = lappend(columnDef, makeConst(TYPEOID,
-								   -1, InvalidOid, -1, typmod, false, true));
+								   -1, InvalidOid, 4, ObjectIdGetDatum(typOid), false, true));
+				columnDef = lappend(columnDef, makeConst(INT4OID,
+								   -1, InvalidOid, 4, Int32GetDatum(typmod), false, true));
 				columnDef = lappend(columnDef, options);
 				columns_list = lappend(columns_list, columnDef);
 				PyMapping_SetItemString(columns_dict, key, column);
@@ -830,17 +831,83 @@ pythonQual(char *operatorname,
 	return qualInstance;
 }
 
+PyObject  *
+getSortKey(MulticornDeparsedSortGroup *key)
+{
+	PyObject *SortKeyClass = getClassString("multicorn.SortKey"),
+			 *SortKeyInstance,
+			 *p_attname,
+			 *p_reversed,
+			 *p_nulls_first,
+			 *p_collate;
+
+	p_attname = PyUnicode_Decode(NameStr(*(key->attname)), strlen(NameStr(*(key->attname))), getPythonEncodingName(), NULL);
+	if (key->reversed)
+		p_reversed = Py_True;
+	else
+		p_reversed = Py_False;
+	if (key->nulls_first)
+		p_nulls_first = Py_True;
+	else
+		p_nulls_first = Py_False;
+	if(key->collate == NULL){
+		p_collate = Py_None;
+		Py_INCREF(p_collate);
+	}
+	else
+		p_collate = PyUnicode_Decode(NameStr(*(key->collate)), strlen(NameStr(*(key->collate))), getPythonEncodingName(), NULL);
+	SortKeyInstance = PyObject_CallFunction(SortKeyClass, "(O,i,O,O,O)",
+			p_attname,
+			key->attnum,
+			p_reversed,
+			p_nulls_first,
+			p_collate);
+	errorCheck();
+	Py_DECREF(p_attname);
+	Py_DECREF(p_collate);
+	Py_DECREF(SortKeyClass);
+	return SortKeyInstance;
+}
+
+MulticornDeparsedSortGroup *
+getDeparsedSortGroup(PyObject *sortKey)
+{
+	MulticornDeparsedSortGroup *md = palloc0(sizeof(MulticornDeparsedSortGroup));
+	PyObject * p_temp;
+	p_temp = PyObject_GetAttrString(sortKey, "attname");
+	md->attname = (Name) strdup(PyUnicode_AS_DATA(p_temp));
+	Py_DECREF(p_temp);
+	p_temp = PyObject_GetAttrString(sortKey, "attnum");
+	md->attnum = (int) PyLong_AsLong(p_temp);
+	Py_DECREF(p_temp);
+	p_temp = PyObject_GetAttrString(sortKey, "is_reversed");
+	md->reversed = PyObject_IsTrue(p_temp);
+	Py_DECREF(p_temp);
+	p_temp = PyObject_GetAttrString(sortKey, "nulls_first");
+	md->nulls_first = PyObject_IsTrue(PyObject_GetAttrString(sortKey, "nulls_first"));
+	Py_DECREF(p_temp);
+	p_temp = PyObject_GetAttrString(sortKey, "collate");
+	if(p_temp == Py_None)
+		md->collate = 0;
+	else
+		md->collate = (Name) strdup(PyUnicode_AS_DATA(p_temp));
+	Py_DECREF(p_temp);
+	return md;
+}
+
 
 /*
  * Execute the query in the python fdw, and returns an iterator.
  */
 PyObject *
-execute(ForeignScanState *node)
+execute(ForeignScanState *node, ExplainState *es)
 {
 	MulticornExecState *state = node->fdw_state;
 	PyObject   *p_targets_set,
 			   *p_quals = PyList_New(0),
-			   *p_iterable;
+			   *p_pathkeys = PyList_New(0),
+			   *p_iterable,
+			   *p_method;
 	ListCell   *lc;
 
 	ExprContext *econtext = node->ss.ps.ps_ExprContext;
@@ -886,22 +953,55 @@ execute(ForeignScanState *node)
 	}
 	/* Transform every object to a suitable python representation */
 	p_targets_set = valuesToPySet(state->target_list);
-	p_iterable = PyObject_CallMethod(state->fdw_instance,
-									 "execute",
-									 "(O,O)",
-									 p_quals,
-									 p_targets_set);
-	errorCheck();
-	if (p_iterable == Py_None)
+
+	foreach(lc, state->pathkeys)
 	{
+		MulticornDeparsedSortGroup *pathkey = (MulticornDeparsedSortGroup *) lfirst(lc);
+		PyObject *python_sortkey = getSortKey(pathkey);
+		PyList_Append(p_pathkeys, python_sortkey);
+		Py_DECREF(python_sortkey);
+	}
+	{
+		PyObject * args,
+				 * kwargs = PyDict_New();
+		if(PyList_Size(p_pathkeys) > 0){
+			PyDict_SetItemString(kwargs, "sortkeys", p_pathkeys);
+		}
+		if(es != NULL){
+			PyObject * verbose;
+			if(es->verbose){
+				verbose = Py_True;
+			} else {
+				verbose = Py_False;
+			}
+			p_method = PyObject_GetAttrString(state->fdw_instance, "explain");
+			args = PyTuple_Pack(2, p_quals, p_targets_set);
+			PyDict_SetItemString(kwargs, "verbose", verbose);
+			errorCheck();
+		} else {
+			p_method = PyObject_GetAttrString(state->fdw_instance, "execute");
+			errorCheck();
+			args = PyTuple_Pack(2, p_quals, p_targets_set);
+			errorCheck();
+		}
+		p_iterable = PyObject_Call(p_method, args, kwargs);
+		errorCheck();
+		Py_DECREF(p_method);
+		Py_DECREF(args);
+		Py_DECREF(kwargs);
+	}
+
+	errorCheck();
+	if (p_iterable == Py_None){
 		state->p_iterator = p_iterable;
 	}
 	else
 	{
 		state->p_iterator = PyObject_GetIter(p_iterable);
 	}
-	Py_DECREF(p_targets_set);
 	Py_DECREF(p_quals);
+	Py_DECREF(p_targets_set);
+	Py_DECREF(p_pathkeys);
 	Py_DECREF(p_iterable);
 	errorCheck();
 	return state->p_iterator;
@@ -1165,6 +1265,11 @@ pythonSequenceToTuple(PyObject *p_value,
 			continue;
 		}
 		p_object = PySequence_GetItem(p_value, j);
+		if(p_object == NULL || p_object == Py_None){
+			nulls[i] = true;
+			values[i] = 0;
+			continue;
+		}
 		resetStringInfo(buffer);
 		values[i] = pyobjectToDatum(p_object, buffer,
 									cinfos[cinfo_idx]);
@@ -1249,7 +1354,7 @@ datumStringToPython(Datum datum, ConversionInfo * cinfo)
 	ssize_t		size;
 	PyObject   *result;
 
-	temp = TextDatumGetCString(datum);
+	temp = datum == 0 ? "?" : TextDatumGetCString(datum);
 	size = strlen(temp);
 	result = PyUnicode_Decode(temp, size, getPythonEncodingName(), NULL);
 	return result;
@@ -1382,7 +1487,7 @@ PyObject *
 datumByteaToPython(Datum datum, ConversionInfo * cinfo)
 {
 	text	   *txt = DatumGetByteaP(datum);
-	char	   *str = VARDATA(txt);
+	char	   *str = txt == NULL ? "?" : VARDATA(txt);
 	size_t		size = VARSIZE(txt) - VARHDRSZ;
 
 #if PY_MAJOR_VERSION >= 3
@@ -1485,7 +1590,7 @@ pathKeys(MulticornPlanState * state)
 		}
 		item = lappend(item, attnums);
 		item = lappend(item, makeConst(INT4OID,
-									 -1, InvalidOid, -1, rows, false, true));
+									 -1, InvalidOid, 4, rows, false, true));
 		result = lappend(result, item);
 		Py_DECREF(p_keys);
 		Py_DECREF(p_cost);
@@ -1496,6 +1601,48 @@ pathKeys(MulticornPlanState * state)
 	return result;
 }
 
+/*
+ * Call the can_sort method from the python implementation. We provide a deparsed
+ * version of the requested fields to sort with all detail as needed (nulls,
+ * collate...), and convert the result to a list of "tuples" (list) of the form:
+ *
+ * - Bitmapset of attnums
+ *
+ * representing the fields that the foreign data wrapper can be sort as
+ * we requested.
+ */
+List *
+canSort(MulticornPlanState * state, List *deparsed)
+{
+	List	   *result = NULL;
+	ListCell   *lc;
+	Py_ssize_t	i;
+	PyObject   *fdw_instance = state->fdw_instance,
+			   *p_pathkeys = PyList_New(0),
+			   *p_sortable;
+
+	foreach(lc, deparsed)
+	{
+		MulticornDeparsedSortGroup *pathkey = (MulticornDeparsedSortGroup *) lfirst(lc);
+		PyObject *python_sortkey = getSortKey(pathkey);
+		PyList_Append(p_pathkeys, python_sortkey);
+		Py_DECREF(python_sortkey);
+	}
+
+	p_sortable = PyObject_CallMethod(fdw_instance, "can_sort", "(O)", p_pathkeys);
+	errorCheck();
+	for (i = 0; i < PySequence_Length(p_sortable); i++)
+	{
+		PyObject   *p_key = PySequence_GetItem(p_sortable, i);
+		MulticornDeparsedSortGroup *md = getDeparsedSortGroup(p_key);
+		result = lappend(result, md);
+		Py_DECREF(p_key);
+	}
+	Py_DECREF(p_pathkeys);
+	Py_DECREF(p_sortable);
+	return result;
+}
+
 PyObject *
 tupleTableSlotToPyObject(TupleTableSlot *slot, ConversionInfo ** cinfos)
 {
@@ -1519,6 +1666,7 @@ tupleTableSlotToPyObject(TupleTableSlot *slot, ConversionInfo ** cinfos)
 		if (isnull)
 		{
 			item = Py_None;
+			Py_INCREF(item);
 		}
 		else
 		{
